#!/usr/bin/perl
#To run do: perl EPSO.pl 1.5 2.5 2.1 0.4 1 5000

use strict;
use Math::Complex;

my $pi=3.141592653;

my $m=$ARGV[0];
my $L=$ARGV[1];
my $gamma=$ARGV[2];
my $T=$ARGV[3];
my $zeta=$ARGV[4];
my $max_N=$ARGV[5];

my $bar_k=2*($m+$L);

printf "m=$m, L=$L, gamma=$gamma, T=$T, zeta=$zeta, t=$max_N, average degree=$bar_k\n";

my $beta=1/($gamma-1);
my @R=();
my %nodeid2label=();
my %nodeid2label_init=();
my %uni_link=();

for (my $n=1; $n <= $max_N; $n++) {

   printf "$n\n";

   my $theta=rand(2*$pi);
   $nodeid2label{$n}{0}=$theta;
   
   $R[$n]=(2/$zeta)*log($n);
   ($R[$n] >=0) or die "Error: $R[$n]\n";

   $nodeid2label{$n}{1}=$R[$n];
   $nodeid2label_init{$n}{1}=$R[$n];

   if($n==1) {
       next;
   }

   #Update node r-coordinates (i.e., move nodes.)
   for (my $j=1; $j < $n ; $j++) {
       $nodeid2label{$j}{1}=$beta*$nodeid2label_init{$j}{1}+(1-$beta)*$R[$n];
   }

   my $I_n=(1/(1-$beta))*(1-$n**(-(1-$beta)));
   my $L_n=(2*$L*(1-$beta)/((2*$beta-1)*((1-$max_N**(-(1-$beta)))**2)))*(($max_N/$n)**(2*$beta-1)-1)*(1-$n**(-(1-$beta)));
   my $m_n=$m+$L_n;

   my $R_n=$R[$n]-(2/$zeta)*log((2*$T/sin($T*$pi))*($I_n/$m_n));

   if($R_n < 0) {
       printf "Warning: negative R_n, setting it to R[n]\n";
       $R_n=$R[$n];
   }
 
   for (my $j=1; $j < $n ; $j++) {

       my $dtheta = abs($nodeid2label{$n}{0}-$nodeid2label{$j}{0});
       if($dtheta > $pi) {
          $dtheta = 2*$pi-$dtheta;
       }
       my $r_a=$nodeid2label{$n}{1};
       my $r_b=$nodeid2label{$j}{1};

       my $d=(1/$zeta)*acosh((cosh($zeta*$r_a)*cosh($zeta*$r_b))-(sinh($zeta*$r_a)*sinh($zeta*$r_b)*cos($dtheta)));
       if( $d < 0 ) {
            die "Error: Negative distance?\n";
       }

       my $p=1/(1+exp((($zeta)/(2*$T))*($d-$R_n)));

       if(rand() <= $p) {
          $uni_link{"$n $j"}=1;
       }

   }

   if ($n==$max_N) {

         open (OUT0,">node_coordinates.txt") or die "Cannot open file node_coordinates.txt\n";
         open (OUT1,">graph.txt") or die "Cannot open file graph.txt\n";

         my %seen=();
         for my $l (keys %uni_link) {
             if  ($l =~ /(\d+) (\d+)/) {
                if(!defined $seen{"$1 $2"}) {
                   print OUT1 "$l\n";
                   $seen{"$1 $2"}=1;
                   $seen{"$2 $1"}=1;
                }
              } else {
                die "Error: bad link\n";
              }
          }
          close(OUT1); 

          for my $n (sort {$a<=>$b} keys %nodeid2label) {
             printf OUT0 "$n $nodeid2label{$n}{0} $nodeid2label{$n}{1}\n";
          }
          close(OUT0);

   }#Endif.

}#Endfor.
