#define _USE_MATH_DEFINES
#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <vector>
#include <iomanip>
#include <omp.h>
#include <sstream>

using namespace std;

class Node {

  private:
  int name, flag;
  long double theta, radius, radius_init, R_i;
  vector<Node*> adjNodeList;

  public:
  Node(int id) {name = id; flag=0;}

  void setAngle(long double angle) {theta=angle;}
  long double getAngle() {return theta;}

  void setRadius(long double r) {radius=r;}
  long double getRadius() {return radius;}

  void setInitRadius(long double r) {radius_init = r;}
  long double getInitRadius() {return radius_init;}

  void set_R(long double R) {R_i = R;}
  long double get_R() {return R_i;}

  void setFlag(int f) {flag = f;}
  int getFlag() {return flag;}

  int getName() {return name;}

  void addAdjNode(Node *adj) {adjNodeList.push_back(adj);}

  vector<Node*> getAdjNodeList() {return adjNodeList;}

  bool checkLink(Node *x) {
    for (int i = 0; i < adjNodeList.size(); i++) {
      if (adjNodeList[i] == x) { return true; }
    }
    return false;
  }
};

class Graph {

  private:
  vector<Node*> nodeList;

  public:
  void addNewNode(Node *nNode) { nodeList.push_back(nNode); }

  Node* findNode(int name) {
    for (int i = 0; i < nodeList.size(); i++) {
      if (nodeList[i]->getName() == name) return nodeList[i];
    }
    return NULL;
  }

  vector<Node*> getNodeList() { return nodeList; }
};

long double Random() {
  long double r = ((long double)rand()/(long double)RAND_MAX);
  return r*2*M_PI;
}

int main(int argc, char *argv[]) {

  long double N = 0, maxDeg = 0, t = 0, m_t = 0, R_t = 0;

  //Change this according to parameters of network to embed.
  long double m = 1.5, L = 1, gamma = 2.1, T = 0.8, zeta = 1;
  
  long double bar_k=2*(m+L);
  long double Beta=1/(gamma-1);

  Graph *G = new Graph();

  Node *u, *v;

  //Read the graph.
  ifstream str(argv[1]);

  if (str.is_open()) {

    string line; 
    int vertex1, vertex2;

    while (getline(str,line)) {
      
      istringstream ss(line);
      string v1, v2;
      ss >> v1;
      ss >> v2;

      char *p1, *p2;
      vertex1 = int(strtol(v1.c_str(), &p1, 10));
      vertex2 = int(strtol(v2.c_str(), &p2, 10));
      
      if (*p1 || *p2) {
        cerr << "Wrong input format: Node IDs must be integers!" << endl;
        exit(1);
      }
      
      u = G->findNode(vertex1);
      if (u == NULL) {
        u = new Node(vertex1);
        G->addNewNode(u);
        N++;
      }

      v = G->findNode(vertex2);
      if (v == NULL) {
        v = new Node(vertex2);
        G->addNewNode(v);
        N++;
      }

      u->addAdjNode(v);
      v->addAdjNode(u);

      if (u->getAdjNodeList().size() > maxDeg) maxDeg = u->getAdjNodeList().size();
      if (v->getAdjNodeList().size() > maxDeg) maxDeg = v->getAdjNodeList().size();
    }

    str.close();
  }

  cout.precision(15);

  srand((unsigned)time(NULL));

  vector<Node*> nodes = G->getNodeList();

  Graph *model = new Graph();

  
  //Embed all nodes with degree >= 1. You can change the 1 in the line below to any other value.
  //For example, changing it to 3, will embed only the nodes will degrees k >= 3.
  for (int i = maxDeg; i >= 1; i--) {
    for (int j = 0; j < nodes.size(); j++) {
      if (nodes[j]->getAdjNodeList().size() == i) {

        t++;
        u = nodes[j];
        long double r_t = (2/zeta)*log(t);
        u->setRadius(r_t);
        u->setInitRadius(r_t);
        model->addNewNode(u);
        vector<Node*> nodes2 = model->getNodeList();
       
        for (int k = 0; k < nodes2.size(); k++) {
          v = nodes2[k];
          if (u != v) {
            v->setRadius(Beta*v->getInitRadius() + (1-Beta)*r_t);
          }
        }

        long double I_t=(1/(1-Beta))*(1-pow(t,-(1-Beta)));
        long double L_t=(2*L*(1-Beta)/((2*Beta-1)*pow((1-pow(N,-(1-Beta))),2)))*(pow(N/t, (2*Beta-1))-1)*(1-pow(t, -(1-Beta)));

        m_t=m+L_t;

        if(t == 1) {
           u->setAngle(Random()); 
           cout << t << " " << u->getName() << " " << u->getAngle() << " " << u->getRadius() << "\n";
           u->setFlag(1);
           continue;
        }

        //To have a more widespread node angular distribution we can ignore the links between the first few nodes for which m_t >= t-1. 
        //This is OK since these nodes can have almost any angle with high probability. If you do not want this, comment the three lines of code below.
        if (m_t >= t-1) {
            u->setFlag(1);
        }

        R_t=r_t-(2/zeta)*log((2*T/sin(T*M_PI))*(I_t/m_t));
           
        if (R_t < 0) {
          cout << "Negative R_t? Setting it to r_t\n";
          R_t=r_t;
        }

        u->set_R(R_t);

        long double step = (long double)1/(t), maxlog_L=-1000000000;
        long double theta=0;

        while (theta <= 2*M_PI) {

          long double theta_t = theta, log_L=0;

          #pragma omp parallel for private(v) reduction(+:log_L)
          for (int k = 0; k < nodes2.size(); k++) {

            v = nodes2[k];   

            if (v != u) {

              long double theta_s = v->getAngle();
              long double r_s = v->getRadius();

              long double dtheta = fabs(theta_t - theta_s);
              if (dtheta > M_PI) {
                dtheta = 2*M_PI - dtheta;
              }

              long double x_st;
              if (dtheta == 0) {
                x_st=fabs(r_t-r_s);
              } else { 
                x_st=(1/zeta)*acosh((cosh(zeta*r_s)*cosh(zeta*r_t))-(sinh(zeta*r_s)*sinh(zeta*r_t)*cos(dtheta)));
              }

              long double P_st = 1/(1+exp((zeta/(2*T))*(x_st-R_t)));

              //If flags are set on both nodes we ignore the links between them (see line 180).
              if ((u->getFlag() == 1) && (v->getFlag() == 1)) {
                  log_L=log_L+log(1-P_st);
                  continue;
              }

              if (u->checkLink(v)) {
                log_L=log_L+log(P_st);
              } else {
                log_L=log_L+log(1-P_st);
              }
            }
          }
      
          if (log_L >= maxlog_L) {  
            u->setAngle(theta_t);
            maxlog_L = log_L;
          }
 
          theta=theta+step;
        }
   
        cout << t << " " << u->getName() << " " << u->getAngle() << " " << u->getRadius() << "\n";

      }//Endif.

    }//Endfor (for all nodes with degree i).

    //The section of the code below runs correction steps as described in the paper. 
    //Put this section into comments if you don't want to run correction steps.
    if ((i==60) || (i==40) || (i==20) || (i==10)) {

      cout << "Running correction step for all nodes with degree greater than " << i << "\n";

      //We repeat each correction step a number of times, equal here to the average node degree (fewer times can also be beneficial.)
      for (int round = 1; round <= bar_k; round++) {

        //cout<<"round "<<round<<"\n";

        vector<Node*> nodes2 = model->getNodeList();

        for (int k = 0; k < nodes2.size(); k++) {

          u = nodes2[k];

          long double step = (long double)1/(t), maxlog_L=-1000000000;
          long double theta=0;

          while (theta <= 2*M_PI) {

            long double theta_u = theta, log_L=0;

            #pragma omp parallel for private(v) reduction(+:log_L)
            for (int l = 0; l < nodes2.size(); l++) {

              v = nodes2[l];

              if (v != u) {

                long double theta_v = v->getAngle();
                long double dtheta = fabs(theta_v - theta_u);
                if (dtheta > M_PI) {
                  dtheta = 2*M_PI - dtheta;
                }

                long double x_vu;
                long double R_init;

                if (u->getInitRadius() > v->getInitRadius()) {//u came after v.  

                  long double r_u=u->getInitRadius();
                  long double r_v=Beta*v->getInitRadius() + (1-Beta)*r_u;
                  if (dtheta == 0) {
                    x_vu=fabs(r_u-r_v);
                  } else {
                    x_vu=(1/zeta)*acosh((cosh(zeta*r_v)*cosh(zeta*r_u))-(sinh(zeta*r_v)*sinh(zeta*r_u)*cos(dtheta)));
                  }
                  R_init=u->get_R();

                } else {

                  long double r_v=v->getInitRadius();
                  long double r_u=Beta*u->getInitRadius() + (1-Beta)*r_v;
                  if (dtheta == 0) {
                    x_vu=fabs(r_u-r_v);
                  } else {
                    x_vu=(1/zeta)*acosh((cosh(zeta*r_v)*cosh(zeta*r_u))-(sinh(zeta*r_v)*sinh(zeta*r_u)*cos(dtheta)));
                  }
                  R_init=v->get_R();

                }

                long double P_uv = 1/(1+exp((zeta/(2*T))*(x_vu-R_init)));

                //If flags are set on both nodes we ignore the links between them (see line 180).
                if ((u->getFlag() == 1) && (v->getFlag() == 1)) {
                   log_L=log_L+log(1-P_uv);
                   continue;
                }

                if (u->checkLink(v)) {
                  log_L=log_L+log(P_uv);
                } else {
                  log_L=log_L+log(1-P_uv);
                }
              }
            }

            if (log_L >= maxlog_L) {
              u->setAngle(theta_u);
              maxlog_L = log_L;
            } 

            theta=theta+step;
          }

        }//Endfor (pick next node).
       
      }//Endfor round.

    }//Endif for correction step.

  }//Endfor (pick next node degree). 

  //Output the node coordinates in the form: node_id angular_coordinate radial_coordinate.
  ofstream out;
  out.precision(15);
  out.open("coordinates_embedding.txt");
  vector<Node*> nodes2 = model->getNodeList();
  for (int j = 0; j < nodes2.size(); j++) {
    u = nodes2[j];
    out << u->getName() << " " << u->getAngle() << " " << u->getRadius() << "\n";
  }
  out.close();

}
